// This file is part of EoS
// Copyright (c) 2009-2016 Thomas Chust
//                         Bayerisches Geoinstitut, Bayreuth
//                         Ludwig-Maximilians-Universität, München
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

namespace EoS.BitmapViewer
open System
open Microsoft.FSharp.Data.UnitSystems.SI.UnitSymbols
open System.Drawing
open System.Windows.Forms
open System.Xml.Linq

type BitmapBox(bitmap:Bitmap, colors:Map<int, int>, p0:float<Pa>, p1:float<Pa>, T0:float<K>, T1:float<K>) as this =
  inherit Control()
  do this.SetStyle(ControlStyles.ResizeRedraw, true)

  member val AnnotationMargin = 40 with get, set
  member val PressureTicks = 11 with get, set
  member val TemperatureTicks = 11 with get, set

  member val Labels = Collections.Generic.List<int * int>()

  override this.OnPaint(evt) =
    let gfx = evt.Graphics
    let mutable rect = this.ClientRectangle

    use sfct =
      new StringFormat(
        Alignment = StringAlignment.Center)
    use sfrm =
      new StringFormat(
        Alignment = StringAlignment.Far,
        LineAlignment = StringAlignment.Center)

    gfx.Clip <- new Region(evt.ClipRectangle)
    gfx.InterpolationMode <- Drawing2D.InterpolationMode.NearestNeighbor

    rect.Inflate(-this.AnnotationMargin, -this.AnnotationMargin)
    gfx.DrawImage(bitmap, rect)

    for i in 0 .. this.PressureTicks-1 do
      let x = rect.Left + i * (rect.Width-1) / (this.PressureTicks-1)
      gfx.DrawLine(Pens.Black, x, rect.Bottom - 3, x, rect.Bottom + 3)
      gfx.DrawLine(Pens.Black, x, rect.Top - 3, x, rect.Top + 3)
      gfx.DrawString(
        sprintf "%.*f" (if p1 - p0 > 1.0e9<Pa> then 0 else 2) <|
        (p0 + float i * (p1 - p0) / float(this.PressureTicks-1)) / 1.0e9<Pa>,
        this.Font, Brushes.Black, float32 x, float32(rect.Bottom + 5), sfct)

    for i in 0 .. this.TemperatureTicks-1 do
      let y = rect.Bottom - i * (rect.Height-1) / (this.TemperatureTicks-1)
      gfx.DrawLine(Pens.Black, rect.Left - 3, y, rect.Left + 3, y)
      gfx.DrawLine(Pens.Black, rect.Right - 3, y, rect.Right + 3, y)
      gfx.DrawString(
        sprintf "%.*f" (if T1 - T0 > 1.0<K> then 0 else 2) <|
        (T0 + float i * (T1 - T0) / float(this.TemperatureTicks-1)) / 1.0<K>,
        this.Font, Brushes.Black, float32(rect.Left - 5), float32 y, sfrm)

    for x, y in this.Labels do
      let color =
        bitmap.GetPixel(x, y)
      let text =
        match colors.TryFind(color.ToArgb()) with
        | Some index -> string index
        | None -> "?"
      let pen, brush =
        if color.GetBrightness() > 0.5f then
          Pens.Black, Brushes.Black
        else
          Pens.White, Brushes.White
      let x, y =
        float32(rect.X + x * (rect.Width-1) / (bitmap.Width-1)),
        float32(rect.Y + y * (rect.Height-1) / (bitmap.Height-1))
      gfx.DrawEllipse(pen, x-2.0f, y-2.0f, 4.0f, 4.0f)
      gfx.DrawString(text, this.Font, brush, x + 1.0f, y + 1.0f)

    base.OnPaint(evt)

  /// Render an SVG representation of the plot.
  member this.ToSVG() =
    let svg = XNamespace.op_Implicit "http://www.w3.org/2000/svg"
    let lnk = XNamespace.op_Implicit "http://www.w3.org/1999/xlink"
    let ns0 = XNamespace.None

    let mutable rect = this.ClientRectangle
    rect.Inflate(-this.AnnotationMargin, -this.AnnotationMargin)

    let image =
      use buf = new IO.MemoryStream()
      bitmap.Save(buf, Imaging.ImageFormat.Png)
      XElement(
        svg + "image",
        XAttribute(ns0 + "id", "bitmap"),
        XAttribute(ns0 + "x", rect.X),
        XAttribute(ns0 + "y", rect.Y),
        XAttribute(ns0 + "width", rect.Width),
        XAttribute(ns0 + "height", rect.Height),
        XAttribute(ns0 + "preserveAspectRatio", "none"),
        XAttribute(
          lnk + "href",
          "data:image/png;base64," +
          Convert.ToBase64String(buf.ToArray())))

    let ticksP =
      XElement(
        svg + "g",
        XAttribute(ns0 + "id", "ticksP"),
        seq {
          for i in 0 .. this.PressureTicks-1 do
            let x = rect.Left + i * (rect.Width-1) / (this.PressureTicks-1)
            yield XElement(
              svg + "g",
              XElement(
                svg + "line",
                XAttribute(ns0 + "x1", x),
                XAttribute(ns0 + "y1", rect.Bottom - 3),
                XAttribute(ns0 + "x2", x),
                XAttribute(ns0 + "y2", rect.Bottom + 3),
                XAttribute(ns0 + "stroke", "black")))
            yield XElement(
              svg + "g",
              XElement(
                svg + "line",
                XAttribute(ns0 + "x1", x),
                XAttribute(ns0 + "y1", rect.Top - 3),
                XAttribute(ns0 + "x2", x),
                XAttribute(ns0 + "y2", rect.Top + 3),
                XAttribute(ns0 + "stroke", "black")),
              XElement(
                svg + "text",
                XAttribute(ns0 + "x", x),
                XAttribute(ns0 + "y", rect.Bottom + 5),
                XAttribute(ns0 + "text-anchor", "middle"),
                XAttribute(ns0 + "dy", 10),
                sprintf "%.*f" (if p1 - p0 > 1.0e9<Pa> then 0 else 2) <|
                (p0 + float i * (p1 - p0) / float(this.PressureTicks-1)) / 1.0e9<Pa>))
        })

    let ticksT =
      XElement(
        svg + "g",
        XAttribute(ns0 + "id", "ticksT"),
        seq {
          for i in 0 .. this.TemperatureTicks-1 do
            let y = rect.Bottom - i * (rect.Height-1) / (this.TemperatureTicks-1)
            yield XElement(
              svg + "g",
              XElement(
                svg + "line",
                XAttribute(ns0 + "x1", rect.Left - 3),
                XAttribute(ns0 + "y1", y),
                XAttribute(ns0 + "x2", rect.Left + 3),
                XAttribute(ns0 + "y2", y),
                XAttribute(ns0 + "stroke", "black")))
            yield XElement(
              svg + "g",
              XElement(
                svg + "line",
                XAttribute(ns0 + "x1", rect.Right - 3),
                XAttribute(ns0 + "y1", y),
                XAttribute(ns0 + "x2", rect.Right + 3),
                XAttribute(ns0 + "y2", y),
                XAttribute(ns0 + "stroke", "black")),
              XElement(
                svg + "text",
                XAttribute(ns0 + "x", rect.Left - 5),
                XAttribute(ns0 + "y", y),
                XAttribute(ns0 + "text-anchor", "end"),
                XAttribute(ns0 + "dy", 5),
                sprintf "%.*f" (if T1 - T0 > 1.0<K> then 0 else 2) <|
                (T0 + float i * (T1 - T0) / float(this.TemperatureTicks-1)) / 1.0<K>))
        })

    let ticked =
      XElement(
        svg + "g",
        XAttribute(ns0 + "id", "ticked"),
        image, ticksP, ticksT)

    let labels =
      XElement(
        svg + "g",
        XAttribute(ns0 + "id", "labels"),
        seq {
          for x, y in this.Labels do
            let color =
              bitmap.GetPixel(x, y)
            let text =
              match colors.TryFind(color.ToArgb()) with
              | Some index -> string index
              | None -> "?"
            let mark =
              if color.GetBrightness() > 0.5f then
                "black"
              else
                "white"
            let x, y =
              rect.X + x * (rect.Width-1) / (bitmap.Width-1),
              rect.Y + y * (rect.Height-1) / (bitmap.Height-1)
            yield XElement(
              svg + "g",
              XElement(
                svg + "circle",
                XAttribute(ns0 + "cx", x),
                XAttribute(ns0 + "cy", y),
                XAttribute(ns0 + "r", 2),
                XAttribute(ns0 + "fill", "none"),
                XAttribute(ns0 + "stroke", mark)),
              XElement(
                svg + "text",
                XAttribute(ns0 + "x", x),
                XAttribute(ns0 + "y", y),
                XAttribute(ns0 + "text-anchor", "start"),
                XAttribute(ns0 + "dx", 3),
                XAttribute(ns0 + "dy", 10),
                XAttribute(ns0 + "fill", mark),
                text))
        })

    XDocument(
      XElement(
        svg + "svg",
        XAttribute(ns0 + "version", "1.1"),
        XAttribute(ns0 + "width", sprintf "%dpx" this.Width),
        XAttribute(ns0 + "height", sprintf "%dpx" this.Height),
        ticked, labels))

  override this.OnClick(evt) =
    match evt with
    | :? MouseEventArgs as evt ->
      let x, y =
        let mutable rect = this.ClientRectangle
        rect.Inflate(-this.AnnotationMargin, -this.AnnotationMargin)
        (evt.X - rect.X) * (bitmap.Width-1) / (rect.Width-1),
        (evt.Y - rect.Y) * (bitmap.Height-1) / (rect.Height-1)
      if 0 <= x && x < bitmap.Width && 0 <= y && y < bitmap.Height then
        this.Labels.Add(x, y)
        this.Refresh()
    | _ ->
      ()
    base.OnClick(evt)
